/***************************************************************************
                          guarddog.cpp  -  description
                             -------------------
    begin                : Thu Feb 10 20:57:36 EST 2000
    copyright            : (C) 2000 by Simon Edwards
    email                : simon@simonzone.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "config.h"

// include files for QT
#include <qlabel.h>
#include <qdir.h>
#include <qstrlist.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qhbox.h>
#include <qbuttongroup.h>
#include <qhbuttongroup.h>
#include <qvbuttongroup.h>

// include files for KDE
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <kstddirs.h>
#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>
#include <klocale.h>
#include <kglobalsettings.h>
#include <unistd.h>
#include <sys/stat.h>

// application specific includes
#include "guarddog.h"
#include "guarddogdoc.h"
#include "resource.h"
#include "protocoldb.h"

///////////////////////////////////////////////////////////////////////////
GuarddogApp::GuarddogApp(const QString &caption) : KDialogBase(KJanusWidget::Tabbed,caption,
        KDialogBase::Help|KDialogBase::Ok|KDialogBase::Apply|KDialogBase::Cancel|KDialogBase::User1,
        KDialogBase::Ok) {

    QString protocollocation;
    QHBox *t;
    QWidget *tmpwidget;
    QLabel *tmplabel;
    QButtonGroup *tmpqbuttongroup;
    int j;
	ProtocolInfo *pinfo;
    QCheckBox *newbox;
    
    tmpwidget = 0;
    waspreviousfirewall = false;
    systemfirewallmodified = false;
    protocollocation = locate("data","guarddog/protocoldatabase");

		// Read in the Protocol Database.
    pdb = new ProtocolDB();
	if(!pdb->read((const char *)(protocollocation))) {
    	KMessageBox::error(0,i18n("An error occured while reading the protocol database.\n\nDetails: \"%1\"").arg(pdb->lastError()));
	}
    
    updatinggui = false;
    aboutus = new KAboutApplication(this);

    setHelp("guarddog/index.html");
    setButtonText(KDialogBase::User1,i18n("About"));
    doc = new GuarddogDoc(pdb);

    generalpage = addVBoxPage(i18n("General"));

        // Internet connection.
    t = new QHBox(generalpage);
    t->setSpacing(spacingHint());
    tmplabel = new QLabel(i18n("Internet connection: "),t);
    t->setStretchFactor(tmpwidget,0);
    internetinterfacecombo = new KComboBox((QWidget *)t);
    internetinterfacecombo->insertItem(i18n("ppp0 - Dialup Interface 0"));
    internetinterfacecombo->insertItem(i18n("ppp1 - Dialup Interface 1"));
    internetinterfacecombo->insertItem(i18n("ppp2 - Dialup Interface 2"));
    internetinterfacecombo->insertItem(i18n("ppp3 - Dialup Interface 3"));
    internetinterfacecombo->insertItem(i18n("eth0 - Ethernet connection 0"));
    internetinterfacecombo->insertItem(i18n("eth1 - Ethernet connection 1"));
    internetinterfacecombo->insertItem(i18n("eth2 - Ethernet connection 2"));
    internetinterfacecombo->insertItem(i18n("eth3 - Ethernet connection 3"));
    internetinterfacecombo->insertItem(i18n("ippp0 - ISDN connection 0"));
    internetinterfacecombo->insertItem(i18n("ippp1 - ISDN connection 1"));
    internetinterfacecombo->insertItem(i18n("ippp2 - ISDN connection 2"));
    internetinterfacecombo->insertItem(i18n("ippp3 - ISDN connection 3"));
    t->setStretchFactor(internetinterfacecombo,1);
    generalpage->setStretchFactor(t,0);
    tmplabel->setBuddy(internetinterfacecombo);
	connect(internetinterfacecombo,SIGNAL(activated(int)),this,SLOT(slotInterfaceBox(int)));

        // Packet blocking group box thing.
    tmpqbuttongroup = new QVButtonGroup(i18n("Packet Blocking"),(QWidget *)generalpage);
    t = new QHBox(tmpqbuttongroup);
    t->setSpacing(spacingHint());
    tmplabel = new QLabel(i18n("Block Method: "),t);
    t->setStretchFactor(tmpwidget,0);
	blockcombo = new KComboBox((QWidget *)t);
    blockcombo->insertItem(i18n("DENY - do not notify source of block"));
    blockcombo->insertItem(i18n("REJECT - notify source of block"));
    t->setStretchFactor(blockcombo,1);
    generalpage->setStretchFactor(t,0);
    tmplabel->setBuddy(blockcombo);
    connect(blockcombo,SIGNAL(activated(int)),this,SLOT(slotBlockMethod(int)));

        // Log blocked packets.
    logblockedpacketscheckbox = new QCheckBox(i18n("Log blocked packets"), (QWidget *)tmpqbuttongroup);
	connect(logblockedpacketscheckbox,SIGNAL(toggled(bool)),this,SLOT(slotLogBlockedPackets(bool)));

        // Ping Group Box thingy
    tmpqbuttongroup = new QHButtonGroup(i18n("Ping"),(QWidget *)generalpage);
        // Allow Ping Check box.
    allowpingcheckbox = new QCheckBox(i18n("Allow Ping Replies"),(QWidget *)tmpqbuttongroup);
    connect(allowpingcheckbox,SIGNAL(toggled(bool)),this,SLOT(slotAllowPingReplies(bool)));
        // Log Ping.
    logpingcheckbox = new QCheckBox(i18n("Log Ping"),(QWidget *)tmpqbuttongroup);
    connect(logpingcheckbox,SIGNAL(toggled(bool)),this,SLOT(slotLogPing(bool)));

        // Advanced settings.
    tmpqbuttongroup = new QHButtonGroup(i18n("Advanced"),(QWidget *)generalpage);
        // Disable firewall.
    disablefirewallcheckbox = new QCheckBox(i18n("Disable firewall"),(QWidget *)tmpqbuttongroup);
    connect(disablefirewallcheckbox,SIGNAL(toggled(bool)),this,SLOT(slotDisableFirewall(bool)));

    rejectauthcheckbox = new QCheckBox(i18n("Always REJECT Auth requests"),(QWidget *)tmpqbuttongroup);
    connect(rejectauthcheckbox,SIGNAL(toggled(bool)),this,SLOT(slotRejectAuth(bool)));
    
        // Just a fakey padding widget that is used for taking up any spare dialog real estate.
    tmpwidget = new QWidget((QWidget *)generalpage);
    generalpage->setStretchFactor(tmpwidget,1);

        // Client Protocols
    clientpage = addVBoxPage(i18n("Client Protocols"));
    new QLabel(i18n("Permitted Outgoing Client Protocols:"),clientpage);
    clientscrollview = new QScrollView(clientpage);
    clientscrollview->viewport()->setBackgroundColor(kapp->palette().active().background());

    t = new QVBox(clientscrollview->viewport());
    t->setSpacing(spacingHint());
    t->setMargin(marginHint());
        // Build our big ass list of client protocol checkboxes.
	j = pdb->clientsList.count();
	if(j!=0) {	// This should always be !0.
		for(pinfo = pdb->clientsList.first(); pinfo!=0; pinfo = pdb->clientsList.next()) {
   		    // Add all the client protocols.
			newbox = new QCheckBox(pinfo->description,(QWidget *)t);
            clientCheckBoxes.append(newbox);
			connect(newbox,SIGNAL(toggled(bool)),this,SLOT(slotProtocolBox(bool)));
		}
	}
    clientscrollview->addChild(t);

        // Server Protocols
    serverpage = addVBoxPage(i18n("Server Protocols"));
    new QLabel(i18n("Permitted Incoming Server Protocols:"),serverpage);
    serverscrollview = new QScrollView(serverpage);
    serverscrollview->viewport()->setBackgroundColor(kapp->palette().active().background());
    t = new QVBox(serverscrollview->viewport());
    t->setSpacing(spacingHint());
    t->setMargin(marginHint());
        // Build our list of server protocol checkboxes.
    j = pdb->serversList.count();
    if(j!=0) {
		for(pinfo = pdb->serversList.first(); pinfo!=0; pinfo = pdb->serversList.next()) {
			newbox = new QCheckBox(pinfo->description,(QWidget *)t);
            serverCheckBoxes.append(newbox);
			connect(newbox,SIGNAL(toggled(bool)),this,SLOT(slotProtocolBox(bool)));
        }
    }
    serverscrollview->addChild(t);

    syncGUI(doc);
	readOptions();
    openDefault();
}

///////////////////////////////////////////////////////////////////////////
GuarddogApp::~GuarddogApp() {
    delete pdb;
    delete aboutus;
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::syncGUI(GuarddogDoc *thisdoc) {
   	char ifbuf[INTERFACE_NAME_LENGTH];
	static char *interfaces[] = {"ppp0","ppp1","ppp2","ppp3","eth0","eth1","eth2","eth3","ippp0","ippp1","ippp2","ippp3",(char *)0};
    QCheckBox *box;
    int i;
    bool disabled;

    updatinggui = true;
    disabled = doc->getDisabled();

        // Checkboxes
    logblockedpacketscheckbox->setChecked(thisdoc->getLogPackets());
    logblockedpacketscheckbox->setDisabled(disabled);

    allowpingcheckbox->setChecked(thisdoc->getAllowPing());
    allowpingcheckbox->setDisabled(disabled);
    logpingcheckbox->setChecked(thisdoc->getLogPing());
    logpingcheckbox->setDisabled(disabled);

        // Block method
    blockcombo->setCurrentItem(doc->getReplyAction());
    blockcombo->setDisabled(disabled);

        // Reject auth
    rejectauthcheckbox->setChecked(thisdoc->getRejectAuth());
    rejectauthcheckbox->setDisabled(disabled);
    
        // Internet interface.
	thisdoc->getInterfaceName(ifbuf);
	for(i=0; interfaces[i]!=0; i++) {
		if(strcmp(ifbuf,interfaces[i])==0) {
			internetinterfacecombo->setCurrentItem(i);
			break;
		}
	}
    internetinterfacecombo->setDisabled(disabled);

		// Update the state of the client protocols list.
	for(i=0; i < (int)clientCheckBoxes.count(); i++) {
		box = clientCheckBoxes.at(i);
		box->setChecked(doc->getClientProtocol(i));
        box->setDisabled(disabled);
	}

		// Update the state of the server protocols list.
	for(i=0; i < (int)serverCheckBoxes.count(); i++) {
		box = serverCheckBoxes.at(i);
		box->setChecked(doc->getServerProtocol(i));
        box->setDisabled(disabled);
	}

    disablefirewallcheckbox->setChecked(disabled);

    updatinggui = false;
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotBlockMethod(int index) {
    if(index==0) {
        doc->setReplyAction(GuarddogDoc::REPLY_ACTION_DENY);
    } else {
        doc->setReplyAction(GuarddogDoc::REPLY_ACTION_REJECT);
    }
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotLogBlockedPackets(bool on) {
    doc->setLogPackets(on);
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotAllowPingReplies(bool on) {
    doc->setAllowPing(on);
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotLogPing(bool on) {
    doc->setLogPing(on);
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotDisableFirewall(bool on) {
    if(updatinggui) {
        return;
    }
    doc->setDisabled(on);
    syncGUI(doc);
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotRejectAuth(bool on) {
    doc->setRejectAuth(on);
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotProtocolBox(bool on) {
	int i;
	QCheckBox *box;

    if(updatinggui) {
        return;
    }

		// Update the state of the Doc client protocol list.
	for(i=0,box=clientCheckBoxes.first(); box!=0; i++,box=clientCheckBoxes.next()) {
		doc->setClientProtocol(i,box->isChecked());
	}

		// Now for Server.
	for(i=0,box=serverCheckBoxes.first(); box!=0; i++,box=serverCheckBoxes.next()) {
		doc->setServerProtocol(i,box->isChecked());
	}
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotInterfaceBox(int index) {
    if(updatinggui) {
        return;
    }
    switch(index) {
		case 0:
			doc->setInterfaceName("ppp0");
			break;
		case 1:
			doc->setInterfaceName("ppp1");
			break;
		case 2:
			doc->setInterfaceName("ppp2");
			break;
		case 3:
			doc->setInterfaceName("ppp3");
			break;
		case 4:
			doc->setInterfaceName("eth0");
			break;
		case 5:
			doc->setInterfaceName("eth1");
			break;
		case 6:
			doc->setInterfaceName("eth2");
			break;
		case 7:
			doc->setInterfaceName("eth3");
			break;
        case 8:
            doc->setInterfaceName("ippp0");
            break;
        case 9:
            doc->setInterfaceName("ippp1");
            break;
        case 10:
            doc->setInterfaceName("ippp2");
            break;
        case 11:
            doc->setInterfaceName("ippp3");
            break;

		default:
			break;
	}
}


///////////////////////////////////////////////////////////////////////////
void GuarddogApp::openFirewall(const QString &filename) {
	QString errorstring;
	int rc;

	rc = doc->openDocument(filename,errorstring);
	if(rc!=GuarddogDoc::OK) {
		KMessageBox::sorry(this,errorstring,i18n("Guarddog"));
	}
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::openDefault() {
	int rc;
    QString filename(SYSTEM_RC_FIREWALL);
    QString errorstring;
	struct stat statbuffer;

	if(stat((const char *)filename,&statbuffer)<0) {
            // There doesn't appear to be a previous Guarddog firewall script.
            // Just warn the user about the ramifications.
        KMessageBox::information(0,i18n("Guarddog was unable to find a Guarddog firewall at %1.\n"
            "This is probably ok, it just means that this is the first time Guarddog has been run on this system.\n"
            "But please be aware that the firewall settings shown may not represent the system's current firewalling configuration.\n"
            "Your Guarddog firewall will take effect once you use the 'Apply' button or exit Guarddog using 'Ok'.").arg(filename));
    } else {
	    rc = doc->openDocument(filename,errorstring);
    	if(rc!=GuarddogDoc::OK) {
                // We were unable to open the guarddog firewall.
            KMessageBox::error(0,i18n("Guarddog was unable to read the file at %1 as being a Guarddog firewall.\n"
            "This probably means that this file in not actually a Guarddog firewall.\n"
            "This is not a problem, but please note that if you exit Guarddog via the 'Ok' button this file will be overwritten.\n"
            "If you do not want this to happen, then immediately quit Guarddog via the 'Cancel' button.\n"
            "Also please be aware that the firewall settings shown may not represent the system's current firewalling configuration.\n\n"
            "(Detailed message \"%2\")").arg(filename).arg(errorstring));
        } else {
            waspreviousfirewall = true;
        }
    }
    syncGUI(doc);
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::saveOptions() {
    KConfig *config;
    config = kapp->config();

    config->setGroup("General");
    config->writeEntry("Geometry", size());
    config->sync();
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::readOptions() {
    KConfig *config;
    config = kapp->config();

    config->setGroup("General");
    QSize size = config->readSizeEntry("Geometry");
    if(!size.isEmpty()) {
        resize(size);
    }
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotOk() {
	QString errorstring;
	int rc;
    QString filename(SYSTEM_RC_FIREWALL);

	rc = doc->saveDocument(filename,errorstring);
    if(rc!=GuarddogDoc::OK) {
		KMessageBox::error(this,i18n("An error occurred while writing the firewall script to disk.\n\n"
            "(Detailed message: \"%1\")").arg(errorstring));
		return;
	}
	if(applyFirewall()) {
        saveOptions();
        accept();
	}
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotApply() {
    applyFirewall();
}

///////////////////////////////////////////////////////////////////////////
bool GuarddogApp::applyFirewall() {
    QString errorstring;
	int rc;
	bool allcool;
	allcool = true;

    if(doc->getDisabled()==false) {
            // Normal firewall apply.
        if(KMessageBox::warningContinueCancel(this,
                i18n("You are about to modify the system's firewall configuration.\n"
                "These changes may disrupt current network connections.\n\n"
                "Do you wish to continue?"),0,i18n("Continue"))==KMessageBox::Continue) {

            rc = doc->applyFirewall(errorstring);
            if(rc!=GuarddogDoc::OK) {
                KMessageBox::error(this,i18n("An error has occurred while trying to modify the system's firewall configuration.\n\n(Detailed message: \"%1\")").arg(errorstring));
                allcool = false;
            } else {
                systemfirewallmodified = true;
                return true;
            }
        }

    } else {
            // If we are to actually disable the firewall because the user has set
            // the Disable checkbox in the advanced section, then we use a different
            // warning question and use a different method to reset the network stack.
        if(KMessageBox::warningContinueCancel(this,
                i18n("You are about to disable the system's firewall.\n"
                "This will allow all network traffic and potentially leave your system vulnerable to attack.\n"
                "Unless you are an advanced user and know what you are doing I recommend that you cancel this action.\n"
                "These changes may also disrupt current network connections.\n\n"
                "Do you wish to continue?"),0,i18n("Continue"))==KMessageBox::Continue) {

            rc = doc->resetSystemFirewall(errorstring);
            if(rc!=GuarddogDoc::OK) {
                KMessageBox::error(this,i18n("An error has occurred while trying to modify the system's firewall configuration.\n\n(Detailed message: \"%1\")").arg(errorstring));
                allcool = false;
            } else {
                systemfirewallmodified = true;
                return true;
            }
        }
    }
 	return allcool;
}

///////////////////////////////////////////////////////////////////////////
void GuarddogApp::slotCancel() {
    QString errorstring;
	int rc;

    if(waspreviousfirewall && systemfirewallmodified) {
            // This is where things become complex.
            // Should we try to restore things to how they were before this program started?
        switch(KMessageBox::warningYesNoCancel(this,
  		    	i18n("The system's firewall settings have been modified.\n\n"
                "Shall I restore them to the previous settings?\n\n"
                "These changes may disrupt current network connections."))) {
                // "Yes, revert to the previous settings."
            case KMessageBox::Yes:
                openDefault();
        		rc = doc->applyFirewall(errorstring);
	        	if(rc!=GuarddogDoc::OK) {
		        	KMessageBox::error(this,i18n("An error has occurred while trying to modify the system's firewall configuration.\n\n(Detailed message: \"%1\")").arg(errorstring));
                } else {
                    saveOptions();
                    accept();
                }
                break;

                // "Just leave the settings alone and piss off!!"
            case KMessageBox::No:
                saveOptions();
                accept();
                break;

                // "Forget I ever pressed the Cancel button."
            case KMessageBox::Cancel:
                break;
        }
    } else {
            // Simple Cancel.
        saveOptions();
        accept();
    }
}

///////////////////////////////////////////////////////////////////////////
//
// About Button
//
void GuarddogApp::slotUser1() {
    aboutus->show();
}
