/***************************************************************************
                          guarddogdoc.h  -  description
                             -------------------
    begin                : Thu Feb 10 20:57:36 EST 2000
    copyright            : (C) 2000 by Simon Edwards
    email                : simon@simonzone.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GUARDDOGDOC_H
#define GUARDDOGDOC_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// include files for QT
#include <qobject.h>
#include <qstring.h>
#include <qlist.h>

#include "protocoldb.h"
class ProtocolDB;
class GuarddogDoc {

public:
    GuarddogDoc(ProtocolDB *database);
    ~GuarddogDoc();

    /** deletes the document's contents */
    void deleteContents();
    /** initializes the document generally */
    bool newDocument();
    /** closes the acutal document */
    void closeDocument();

	enum ReturnCode {OK,READ_ERROR,WRITE_ERROR,PARSE_ERROR};

    /** loads the document by filename and format and emits the updateViews() signal */
    int openDocument(const QString &filename, QString &errorstring);
    /** saves the document under filename and format.*/
    int saveDocument(const QString &filename, QString &errorstring);
	/** Runs the firewall and updates the current ipchains rules. */
	int applyFirewall(QString &errorstring);

    int resetSystemFirewall(QString &errorstring);

	/** Runs the given firewall file. */
	int runFirewall(const QString &filename,QString &errorstring);
    int writeFirewall(FILE *fhandle,QString &errorstring);
	int readFirewall(FILE *fhandle,QString &errorstring);

	void setClientProtocol(int index,bool state);
	bool getClientProtocol(int index);

	bool getServerProtocol(int index);
	void setServerProtocol(int index,bool state);

    void setAllowPing(bool on);
    bool getAllowPing();

    void setLogPing(bool on);
    bool getLogPing();

    void setLogPackets(bool on);
	bool getLogPackets();

    void setInterfaceName(const char *name);
	void getInterfaceName(char *name);

    void setDisabled(bool on);
    bool getDisabled();

	enum REPLY_ACTION {REPLY_ACTION_DENY,REPLY_ACTION_REJECT};
    int getReplyAction();
    void setReplyAction(REPLY_ACTION action);

    void setLocalPortRange(uint low,uint high);

    void setRejectAuth(bool on);
    bool getRejectAuth();
        
private:
	bool *allowedClientProtocols;
	bool *allowedServerProtocols;
	bool logPackets;
	bool allowPing;
	bool logPing;
    bool disabled;
    uint localPortLow;
    uint localPortHigh;
    bool rejectAuth;
    
#define INTERFACE_NAME_LENGTH	16
	char interfacename[INTERFACE_NAME_LENGTH];
	REPLY_ACTION replyAction;
	bool fprintExpandRule(FILE *fhandle,const char *format,ProtocolInfo::PortSpec *leftspec,
		ProtocolInfo::PortSpec *rightspec);
	bool fprintExpandRightPorts(FILE *fhandle,const char *format,const char *left,
		ProtocolInfo::PortSpec *spec);
	void factoryDefaults();
    void guessInternetInterface(char *dest);
    ProtocolDB *pdb;
};

#endif // GUARDDOGDOC_H
