/***************************************************************************
                          protocolinfo.cpp  -  description
                             -------------------
    begin                : Thu Feb 10 2000
    copyright            : (C) 2000 by Simon Edwards
    email                : simon@simonzone.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include "protocolinfo.h"
#include "stringparse.h"

///////////////////////////////////////////////////////////////////////////
ProtocolInfo::ProtocolInfo(){
    name = 0;
    description = 0;
    allowInTCP = 0;
    allowInTCPSource = 0;
    allowOutTCP = 0;
    allowOutTCPSource = 0;
    allowInUDP = 0;
    allowInUDPSource = 0;
    allowOutUDP = 0;
    allowOutUDPSource = 0;
}

///////////////////////////////////////////////////////////////////////////
ProtocolInfo::~ProtocolInfo(){
    destroyPortSpec(allowInTCP);
    destroyPortSpec(allowInTCPSource);
    destroyPortSpec(allowOutTCP);
    destroyPortSpec(allowOutTCPSource);
    destroyPortSpec(allowInUDP);
    destroyPortSpec(allowInUDPSource);
    destroyPortSpec(allowOutUDP);
    destroyPortSpec(allowOutUDPSource);
    
    delete name;
    delete description;
}

///////////////////////////////////////////////////////////////////////////
void ProtocolInfo::destroyPortSpec(PortSpec *x) {
	PortSpec *p,*q;
	
	p = x;
	while(p) {
		q = p->next;
		delete p;
		p = q;
	}
}

///////////////////////////////////////////////////////////////////////////
bool ProtocolInfo::parseLine(StringParse &parser) {
#define WORD_BUFFER_SIZE    512
    char wordbuffer[WORD_BUFFER_SIZE];

        // Read the name.
    if(!parser.getString(wordbuffer,WORD_BUFFER_SIZE)) {
            // Should do some kind of error string.
        return false;
    }
    name = new char[strlen(wordbuffer)+1];
    strcpy(name,wordbuffer);
    
    parser.skipWhite();
    
        // Read in the In TCP port list.
    if(!parsePortSpec(parser,&allowInTCP)) {
        return false;    
    }
    parser.skipWhite();
    
        // Read in the In TCP source port list.    
    if(!parsePortSpec(parser,&allowInTCPSource)) {
        return false;    
    }
    parser.skipWhite();
        
        // Read in the Out TCP port list.    
    if(!parsePortSpec(parser,&allowOutTCP)) {
        return false;    
    }
    parser.skipWhite();
        
        // Read in the Out TCP port list.    
    if(!parsePortSpec(parser,&allowOutTCPSource)) {
        return false;    
    }
    parser.skipWhite();

        // Read in the In UDP port list.
    if(!parsePortSpec(parser,&allowInUDP)) {
        return false;    
    }
    parser.skipWhite();
    
        // Read in the In UDP source port list.    
    if(!parsePortSpec(parser,&allowInUDPSource)) {
        return false;    
    }
    parser.skipWhite();
        
        // Read in the Out UDP port list.    
    if(!parsePortSpec(parser,&allowOutUDP)) {
        return false;    
    }
    parser.skipWhite();
        
        // Read in the Out UDP port list.    
    if(!parsePortSpec(parser,&allowOutUDPSource)) {
        return false;    
    }
    parser.skipWhite();
    
        
        // Read the description.
    if(!parser.getString(wordbuffer,WORD_BUFFER_SIZE)) {
            // Should do some kind of error string.
        return false;
    }
    description = new char[strlen(wordbuffer)+1];
    strcpy(description,wordbuffer);
    
    return true;
}

///////////////////////////////////////////////////////////////////////////
bool ProtocolInfo::parsePortSpec(StringParse &parser,PortSpec **x) {
#define TEMP_BUFFER_SIZE    256
    char tempbuffer[TEMP_BUFFER_SIZE];
    uint lport,rport;
    char c;
    StringParse sp;
    PortSpec *tempportspec;
    PortSpec **currentptr;
    
    currentptr = x;
    if(parser.getWord(tempbuffer,TEMP_BUFFER_SIZE)) {
        sp.setStr(tempbuffer);
        while(!sp.endOfString()) {
                
                // Lets see if we have a port spec. '-' means no ports.
            sp.mark();
            if(!sp.getChar(c)) {
                return true;
            }
            
            switch(c) {
                case '-':
                    return true;
                        
                case '*':
                    // Any port.
                    tempportspec = new PortSpec();
                    tempportspec->next = 0;
                    tempportspec->portStart = 0;
                    tempportspec->portEnd = 65535;
                    tempportspec->localPorts = false;
                    *currentptr = tempportspec;
                    currentptr = &((*currentptr)->next);
                    break;
    			    
			    case '+':
                    // Only unprivliged ports.
                    tempportspec = new PortSpec();
                    tempportspec->next = 0;
                    tempportspec->portStart = 1024;
                    tempportspec->portEnd = 65535;
                    tempportspec->localPorts = false;
                    *currentptr = tempportspec;
                    currentptr = &((*currentptr)->next);
    				break;
        			
    			case 'l':
    			case 'L':
                    // Only local ports.
                    tempportspec = new PortSpec();
                    tempportspec->next = 0;
                    tempportspec->portStart = 1024; // These will be changed later.
                    tempportspec->portEnd = 65535;
                    tempportspec->localPorts = true;
                    *currentptr = tempportspec;
		    		return true;
    		    		
			    default:
			        sp.rollback();
                        // Encountered something wierd.
                    if(!sp.getUint(lport)) {
                        // TODO: We should really look up any port names.
                        return false;
                    }
                        
                        // Try to detect a port range.
                    if(sp.compareSubword(":")) {
                        sp.getChar(c);
                        if(!sp.getUint(rport)) {
                            return false;
                        }
                            // This is a port range.
                        tempportspec = new PortSpec();
                        tempportspec->next = 0;
                        tempportspec->portStart = lport;
                        tempportspec->portEnd = rport;
                        *currentptr = tempportspec;
                        currentptr = &((*currentptr)->next);
                        break;
                    }
                        
                        // It was just a lone number, add it to the list.
                    tempportspec = new PortSpec();
                    tempportspec->next = 0;
                    tempportspec->portStart = lport;
                    tempportspec->portEnd = lport;
                    *currentptr = tempportspec;
                    currentptr = &((*currentptr)->next);
                    break;
            }

            if(!sp.getChar(c)) {
                return true;
            }            
            if(c!=',') {
                return false;
            }
        }
            // End of string hit.
        return true;
    } else {
        return false;
    }
    
}

///////////////////////////////////////////////////////////////////////////
void ProtocolInfo::setLocalPortRange(uint low, uint high) {
    setPortRangeSpec(allowInTCP,low,high);
    setPortRangeSpec(allowInTCPSource,low,high);
    setPortRangeSpec(allowOutTCP,low,high);
    setPortRangeSpec(allowOutTCPSource,low,high);
    setPortRangeSpec(allowInUDP,low,high);
    setPortRangeSpec(allowInUDPSource,low,high);
    setPortRangeSpec(allowOutUDP,low,high);
    setPortRangeSpec(allowOutUDPSource,low,high);
}
///////////////////////////////////////////////////////////////////////////
void ProtocolInfo::setPortRangeSpec(PortSpec *spec,uint low,uint high) {
    while(spec) {
        if(spec->localPorts) {
            spec->portStart = low;
            spec->portEnd = high;
        }
        spec = spec->next;
    }
}