/***************************************************************************
                          stringparse.cpp  -  description
                             -------------------
    begin                : Tue Feb 1 2000
    copyright            : (C) 2000 by Simon Edwards
    email                : simon@simonzone.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stringparse.h"
#include <ctype.h>
#include <string.h>

///////////////////////////////////////////////////////////////////////////
StringParse::StringParse() {
    str = (char *)"";
    pos = 0;
    markpos = 0;
}

///////////////////////////////////////////////////////////////////////////
StringParse::~StringParse() {
}

///////////////////////////////////////////////////////////////////////////
void StringParse::setStr(char *line) {
    str = line;
    pos = 0;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::compareWord(char *word) {
    int n;
    
    n = pos;
    while(str[n]!=0 && *word!=0 && str[n]==*word) {
        n++;
        word++;
    }
    if(str[n]==0) {
        if(*word==0) {
            return true;
        }
        return 0;
    } else {
        if(isspace(str[n])) {
            return true;
        } else {
            return false;
        }
    }
}

///////////////////////////////////////////////////////////////////////////
void StringParse::skipWhite() {
    while(str[pos]!=0 && isspace(str[pos])) {
        pos++;
    }
}

///////////////////////////////////////////////////////////////////////////
void StringParse::skipWord() {
    while(str[pos]!=0 && !isspace(str[pos])) {
        pos++;
    }
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::compareSubword(char *subword) {    
    int n;
    n = pos;
    while(str[n]!=0 && *subword!=0 && str[n]==*subword) {
        n++;
        subword++;
    }
    if(*subword==0) {
        return true;
    }
    return false;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::getUint(uint &n) {
    uint x;
    int used;
    
    if(sscanf(str+pos,"%u%n",&x,&used)!=1) {
        return false;
    }
    n = x;
    pos += used;
    return true;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::getChar(char &c) {

    if(str[pos]==0) {
        return false;
    }
    c = str[pos];
    pos++;
    return true;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::getIP(struct in_addr &IP) {
    uint byte;
    int i;
    int posbackup;
    
    posbackup = pos;
            
    IP.s_addr = 0;
    for(i=0; i<4; i++) {
        if(!getUint(byte)) {
            pos = posbackup;
            return false;
        }    
        if(byte>255) {
            pos = posbackup;
            return false;
        }
        IP.s_addr <<= 8;
        IP.s_addr |= byte;
        if(i!=3) {
            if(!compareSubword(".")) {
                pos = posbackup;
                return false;
            }
            skipChars(1);
        }
    }
    return true;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::getHexUint(uint &n) {
    uint x;
    int used;
    
    if(sscanf(str+pos,"0x%x%n",&x,&used)!=1) {
        return false;
    }
    n = x;
    pos += used;
    return true;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::skipChars(int n) {
    while(n && str[pos]!=0) {
        pos++;        
        n--;
    }
    return n==0;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::getWord(char *buf,int bufsize) {
    int i;
    
    i=0;
        // Count how much we want to copy
    while(!isspace(str[pos+i]) && str[pos+i]!=0) {
        i++;
    }
        // No word or not enough space.
    if(i==0 || i>=bufsize) {
        return false;
    }
    
    memcpy(buf,str+pos,i);
    buf[i] = 0;
    pos += i;
    return true;    
}

///////////////////////////////////////////////////////////////////////////
char *StringParse::currentPosition() {
    return str+pos;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::endOfString() {
    return str[pos]==0;
}

///////////////////////////////////////////////////////////////////////////
bool StringParse::getString(char *buffer,int buffersize) {
    int backuppos;
    int i;
    char c;
    char quote;
    
    quote = 0;
    backuppos = pos;

    if(getChar(c)) {
        if(c=='\'' || c=='"') {
            quote = c;
            i=0;
                // Count how much we want to copy
            while(str[pos+i]!=quote && str[pos+i]!=0) {
                i++;
            }
            
            if(str[pos+i]!=quote) {
                    // We must have non-terminated string.
                return false;
            }
            
                // No word or not enough space.
            if(i>=buffersize) {
                return false;
            }
            if(i!=0) {
                memcpy(buffer,str+pos,i);
            }
            buffer[i] = 0;
            pos += i;
            pos++;
            return true;    
        } else {
                // Do a get word when we don't have quotes.
            pos = backuppos;
            return getWord(buffer,buffersize);
        }
        
    } else {
        pos = backuppos;
        return false;
    }
}

///////////////////////////////////////////////////////////////////////////
void StringParse::mark() {
    markpos = pos;
}

///////////////////////////////////////////////////////////////////////////
void StringParse::rollback() {
    pos = markpos;
}
